# resolvent_test.py
#
# Test that attempts to isolate a bug in the application
# of the resolvent R(z) = (zB - A)**(-1).


from netgen.geom2d import unit_square, SplineGeometry
import ngsolve as ngs
import numpy as np


def make_unit_square_mesh(maxh=0.2):
    return ngs.Mesh(unit_square.GenerateMesh(maxh=maxh))


def make_fiber_pml_mesh(maxh=2.0, rcore=1.0, rclad=2.0, rout=20.0):
    geo = SplineGeometry()

    return ngs.Mesh(geo.GenerateMesh())


def make_spaces_and_forms(mesh, p, dirichlet):
    X = ngs.H1(mesh, p, dirichlet=dirichlet, complex=True)

    # Test and trial functions.
    u, v = X.TnT()

    # Create the bilinear forms.
    a = ngs.BilinearForm(X)
    a += (ngs.grad(u) * ngs.grad(v)) * ngs.dx

    b = ngs.BilinearForm(X)
    b += (u * v) * ngs.dx

    # Assemble the bilinear forms.
    with ngs.TaskManager():
        a.Assemble()
        b.Assemble()

    return X, a, b


def make_resolvents(X, a, b, z, inverse=''):
    A = a.mat
    B = b.mat

    R = []

    for k in range(len(z)):

        with ngs.TaskManager():
            # Create the inverse of the resolvent matrix.
            rinv = A.CreateMatrix()
            rinv.AsVector().data = complex(z[k]) * B.AsVector() - A.AsVector()

            # Compute the resolvent proper and add it to the list.
            r = rinv.Inverse(X.FreeDofs(), inverse=inverse)

        R.append(r)

    return R


def apply_resolvents(X, R, f, hermitian_conjugate=True):
    # Create the solution, and zero it out.
    u = ngs.GridFunction(X)
    u.Set(ngs.CoefficientFunction(0))

    if hermitian_conjugate:
        for k in range(len(R)):
            u.vec.data += R[k].H * f.vec
    else:
        for k in range(len(R)):
            u.vec.data += R[k] * f.vec
 
    return u


def resolvent_test(maxh, p, inverse='', meshtype='square',
                   hermitian_conjugate=True):
    """
    resolvent_test()

    Method that computes u = \sum_{i=0}^{N-1} (z[i]*B - A)**(-1) * f
    for some vector f and array of complex numbers
    z.

    INPUTS:

    maxh     - the maximum element diameter for the mesh.
    p        - polynomial degree for the finite element space
    inverse  - the type of driver to use to compute the inverse, can be
               one of the following (documentation is provided from the
               python command line):

             '' - the default, which calls on pardiso to compute the
                  inverse of z*B - A (since this install of NGSolve was
                  compiled with -DUSE_MKL=ON and -DUSE_UNFPACK=ON)
             'sparsecholesky' - internal solver of NGSolve for symmetric
                                matrices
             'umfpack'        - solver by Suitesparse/UMFPACK (if NGSolve
                                was configured with USE_UMFPACK=ON)
             'pardiso'        - PARDISO, either provided by libpardiso
                                (USE_PARDISO=ON) or Intel MKL (USE_MKL=ON).
                                If neither Pardiso nor Intel MKL was linked
                                at compile-time, NGSolve will look for
                                libmkl_rt in LD_LIBRARY_PATH (Unix) or PATH
                                (Windows) at run-time. 
             

    meshtype - generate either a square or circular mesh
    hermitian_conjugate - If True, apply the hermitian conjugate of the
                          resolvents (z[i]*B - A)**(-1) to the right-
                          hand-side function f. Otherwise, apply the
                          resolvent as-is to the right-hand-side f.

    OUTPUTS:

    None
    """

    # Create the mesh.
    if meshtype == 'square':
        print('Creating squre mesh')
        mesh = make_unit_square_mesh(maxh=maxh)
        dirichlet = '.*'
    else:
        print('Creating circular mesh')
        mesh = make_fiber_pml_mesh(maxh=maxh)
        dirichlet = 'outer'

    # Create the finite element space and forms.
    print('Creating finite element space and forms')
    X, a, b = make_spaces_and_forms(mesh=mesh, p=p, dirichlet=dirichlet)

    # Give a set of four "quadrature points."
    print('Creating quadrature points')
    z = np.exp(1j * np.linspace(0, 2*np.pi, 4, endpoint=False))

    # Make the resolvents.
    print('Making the resolvent matrices using ' + 
          'inverse =\'{0}\''.format(inverse))
    R = make_resolvents(X, a, b, z, inverse=inverse)

    # Create an arbitrary right-hand-side.
    print('Creating a grid function for the right-hand-side')
    f = ngs.GridFunction(X)
    f.Set(ngs.x)

    # Apply the resolvents.
    print('Calling apply_resolvents() with ' + 
          'hermitian_conjugate = {0}'.format(hermitian_conjugate))
    u = apply_resolvents(X, R, f)
    print('Returning from apply_resolvents()')


if __name__ == '__main__':
    maxh = 0.2
    p = 1
    inverse = ''
    meshtype = 'square'
    hermitian_conjugate = False

    resolvent_test(maxh, p, inverse, meshtype, hermitian_conjugate)
